/*
 * ProFTPD: mod_zeroconf -- a module for advertising ProFTPD via the
 *                          Zeroconf protocols
 *
 * Copyright (c) 2006 TJ Saunders
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * As a special exemption, TJ Saunders and other respective copyright holders
 * give permission to link this program with OpenSSL, and distribute the
 * resulting executable, without including the source code for OpenSSL in the
 * source distribution.
 *
 * This is mod_zeroconf, contrib software for proftpd 1.3.x and above.
 * For more information contact TJ Saunders <tj@castaglia.org>.
 *
 * $Id: mod_zeroconf.c,v 1.2 2006/05/19 22:42:10 tj Exp tj $
 */

#include "conf.h"
#include "mod_ctrls.h"

#define MOD_ZEROCONF_VERSION		"mod_zeroconf/0.1"

/* Make sure the version of proftpd is as necessary. */
#if PROFTPD_VERSION_NUMBER < 0x0001030001
# error "ProFTPD 1.3.0rc1 or later required"
#endif

/* Currently only supports Bonjour. */
#ifndef DARWIN8
# error "Mac OS X 10.3 or later required"
#endif

#include <dns_sd.h>

extern xaset_t *server_list;

module zeroconf_module;

static int zeroconf_engine = FALSE;
static char *zeroconf_log = NULL;
static int zeroconf_logfd = -1;
static pool *zeroconf_pool = NULL;
static array_header *zeroconf_svc_list = NULL;
static array_header *zeroconf_txt_list = NULL;

/* Support routines
 */

static const char *zeroconf_strerror(DNSServiceErrorType code) {
  switch (code) {
    case kDNSServiceErr_NoError:
      return "No error";

    case kDNSServiceErr_Unknown:
      return "Unknown error";

    case kDNSServiceErr_NoSuchName:
      return "No such name";

    case kDNSServiceErr_NoMemory:
      return "Out of memory";

    case kDNSServiceErr_BadParam:
      return "Invalid argument";

    case kDNSServiceErr_BadReference:
      return "Invalid reference";

    case kDNSServiceErr_BadState:
      return "Bad state";

    case kDNSServiceErr_BadFlags:
      return "Bad flags";

    case kDNSServiceErr_Unsupported:
      return "Operation unsupported";

    case kDNSServiceErr_NotInitialized:
      return "Not initialized";

    case kDNSServiceErr_AlreadyRegistered:
      return "Already registered";

    case kDNSServiceErr_NameConflict:
      return "Name conflict";

    case kDNSServiceErr_Invalid:
      return "Invalid operation";

    case kDNSServiceErr_Firewall:
      return "Firewall issues";

    case kDNSServiceErr_Incompatible:
      return "Incompatible";

    case kDNSServiceErr_BadInterfaceIndex:
      return "Bad interface index";

    case kDNSServiceErr_Refused:
      return "Registration refused";

    case kDNSServiceErr_NoSuchRecord:
      return "No such record";

    case kDNSServiceErr_NoAuth:
      return "No authorization";

    case kDNSServiceErr_NoSuchKey:
      return "No such key";

    case kDNSServiceErr_NATTraversal:
      return "NAT traversal";

    case kDNSServiceErr_DoubleNAT:
      return "Double NAT";

    case kDNSServiceErr_BadTime:
      return "Bad time";

    default:
      return "(unknown)";
  }

  return NULL;
}

static DNSServiceRef *zeroconf_get_handle(void) {
  DNSServiceRef *h = push_array(zeroconf_svc_list);
  return h;
}

static const char *zeroconf_get_name(server_rec *s) {
  config_rec *c;

  c = find_config(s->conf, CONF_PARAM, "ZeroconfInfo", FALSE);
  if (c) {
    return (const char *) c->argv[0];
  }

  return s->ServerName;
}

static TXTRecordRef *zeroconf_get_info(server_rec *s) {
  config_rec *c;

  c = find_config(s->conf, CONF_PARAM, "ZeroconfInfo", FALSE);
  if (c != NULL &&
      c->argv[1] != NULL) {
    char *info = c->argv[1], *str;

    /* Assume that the 'info' string is a colon-delimited list of key/value
     * pairs, e.g.:
     *
     *  anonymous:ssl=TLSv1
     *
     * If so, create a TXTRecordRef object and populate it with the info.
     */

     TXTRecordRef *t = push_array(zeroconf_txt_list);
     TXTRecordCreate(t, 0, NULL);

     str = info;
     while (str != NULL &&
            strlen(str) > 0) {
       char *tmp;

       pr_signals_handle();

       tmp = strchr(str, ':');
       if (tmp) {
         char *tmp2;

         *tmp = 0;
         tmp2 = strchr(str, '=');
         if (tmp2) {
           char *k = str, *v = tmp2 + 1;
           DNSServiceErrorType res;

           *tmp2 = 0;
           res = TXTRecordSetValue(t, k, strlen(v), v);
           if (res != kDNSServiceErr_NoError) {
             (void) pr_log_writefile(zeroconf_logfd, MOD_ZEROCONF_VERSION,
               "error adding key '%s', value '%s' to TXT record: %s", k, v,
               zeroconf_strerror(res));
           } else {
             (void) pr_log_writefile(zeroconf_logfd, MOD_ZEROCONF_VERSION,
               "added key '%s', value '%s' to TXT record", k, v);
           }

           *tmp2 = '=';

         } else {
           DNSServiceErrorType res = TXTRecordSetValue(t, str, 0, NULL);
           if (res != kDNSServiceErr_NoError) {
             (void) pr_log_writefile(zeroconf_logfd, MOD_ZEROCONF_VERSION,
               "error adding key '%s' to TXT record: %s", str,
               zeroconf_strerror(res));
           } else {
             (void) pr_log_writefile(zeroconf_logfd, MOD_ZEROCONF_VERSION,
               "added TXT record for key '%s'", str);
           }
         }

         *tmp = ':';
         str = tmp + 1;
         tmp = strchr(str, ':');

       } else {
         char *tmp2 = strchr(str, '=');
         if (tmp2) {
           char *k = str, *v = tmp2 + 1;
           DNSServiceErrorType res;

           *tmp2 = 0;
           res = TXTRecordSetValue(t, k, strlen(v), v);
           if (res != kDNSServiceErr_NoError) {
             (void) pr_log_writefile(zeroconf_logfd, MOD_ZEROCONF_VERSION,
               "error adding key '%s', value '%s' to TXT record: %s", k, v, 
               zeroconf_strerror(res));

           } else {
             (void) pr_log_writefile(zeroconf_logfd, MOD_ZEROCONF_VERSION,
               "added key '%s', value '%s' to TXT record", k, v);
           }

         } else {
           DNSServiceErrorType res = TXTRecordSetValue(t, str, 0, NULL);
           if (res != kDNSServiceErr_NoError) {
             (void) pr_log_writefile(zeroconf_logfd, MOD_ZEROCONF_VERSION,
               "error adding key '%s' to TXT record: %s", str,
               zeroconf_strerror(res));

           } else {
             (void) pr_log_writefile(zeroconf_logfd, MOD_ZEROCONF_VERSION,
               "added key '%s' to TXT record", str);
           }
         }

         str = NULL;
       }
     }
 
     return t;
  }

  return NULL;
}

static void zeroconf_callback(DNSServiceRef handle, DNSServiceFlags flags,
    DNSServiceErrorType code, const char *name, const char *regtype,
    const char *domain, void *user_data) {

  if (code == kDNSServiceErr_NoError) {
    (void) pr_log_writefile(zeroconf_logfd, MOD_ZEROCONF_VERSION,
      "successfully registered name '%s' (domain '%s')", name, domain);
    return;
  }

  (void) pr_log_writefile(zeroconf_logfd, MOD_ZEROCONF_VERSION,
    "error registering name '%s' (domain '%s'): %s", name, domain,
    zeroconf_strerror(code));
}

static void zeroconf_register(void) {
  DNSServiceErrorType res;
  server_rec *s;

  /* Iterate through the configured servers, looking for info that they
   * want to share.
   */
  for (s = (server_rec *) server_list->xas_list; s; s = s->next) {
    DNSServiceRef *h = zeroconf_get_handle();
    const char *name = zeroconf_get_name(s);
    TXTRecordRef *t = zeroconf_get_info(s);
    
    res = DNSServiceRegister(h, 0, kDNSServiceInterfaceIndexAny,
      name, "_ftp._tcp", NULL, NULL, ntohs(s->ServerPort),
      t ? TXTRecordGetLength(t) : 0,
      t ? TXTRecordGetBytesPtr(t) : NULL, zeroconf_callback, NULL);
    if (res != kDNSServiceErr_NoError) {
      (void) pr_log_writefile(zeroconf_logfd, MOD_ZEROCONF_VERSION,
        "error registering name '%s': %s", name, zeroconf_strerror(res));
      return;
    }

    res = DNSServiceProcessResult(*h);
    if (res != kDNSServiceErr_NoError) {
      (void) pr_log_writefile(zeroconf_logfd, MOD_ZEROCONF_VERSION,
        "error processing results of '%s' registration: %s", name,
        zeroconf_strerror(res));
      return;
    }
  }
}

static void zeroconf_unregister(void) {
  if (zeroconf_txt_list->nelts > 0) {
    register unsigned int i;
    TXTRecordRef *ts = (TXTRecordRef *) zeroconf_txt_list->elts;

    for (i = 0; i < zeroconf_txt_list->nelts; i++) {
      TXTRecordRef *t = &(ts[i]);
      TXTRecordDeallocate(t);
    }
  }

  if (zeroconf_svc_list->nelts > 0) {
    register unsigned int i;
    DNSServiceRef *rs = (DNSServiceRef *) zeroconf_svc_list->elts;

    for (i = 0; i < zeroconf_svc_list->nelts; i++) {
      DNSServiceRef *r = &(rs[i]);
      DNSServiceRefDeallocate(*r);
    }
  }
}

/* Configuration handlers
 */

/* usage: ZeroconfEngine on|off */
MODRET set_zeroconfengine(cmd_rec *cmd) {
  int bool = -1;

  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT);

  bool = get_boolean(cmd, 1);
  if (bool == -1)
    CONF_ERROR(cmd, "expected Boolean parameter");

  zeroconf_engine = bool;
  return HANDLED(cmd);
}

/* usage: ZeroconfInfo name [info] */
MODRET set_zeroconfinfo(cmd_rec *cmd) {
  config_rec *c;

  if (cmd->argc-1 < 1 || cmd->argc-1 > 2)
    CONF_ERROR(cmd, "wrong number of parameters");

  CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL);

  c = add_config_param(cmd->argv[0], 2, NULL, NULL);
  c->argv[0] = pstrdup(c->pool, cmd->argv[1]);
  if (cmd->argc-1 == 2)
    c->argv[1] = pstrdup(c->pool, cmd->argv[2]);

  return HANDLED(cmd);
}

/* usage: ZeroconfLog path|"none" */
MODRET set_zeroconflog(cmd_rec *cmd) {
  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT);

  if (pr_fs_valid_path(cmd->argv[1]) < 0)
    CONF_ERROR(cmd, "not a valid path");

  zeroconf_log = pstrdup(cmd->server->pool, cmd->argv[1]);
  return HANDLED(cmd);
}

/* Event handlers
 */

static void zeroconf_exit_ev(const void *event_data, void *user_data) {
  zeroconf_unregister();
}

static void zeroconf_postparse_ev(const void *event_data, void *user_data) {
  if (!zeroconf_engine) {
    destroy_pool(zeroconf_pool);
    zeroconf_pool = NULL;
    zeroconf_svc_list = NULL;
    zeroconf_txt_list = NULL;

    return;
  }

  /* Open the ZeroconfLog, if requested. */
  if (zeroconf_log &&
      strcasecmp(zeroconf_log, "none") != 0) {
    if (pr_log_openfile(zeroconf_log, &zeroconf_logfd, 0666) < 0)
      pr_log_debug(DEBUG2, MOD_ZEROCONF_VERSION
        ": error opening ZeroconfLog '%s': %s", zeroconf_log, strerror(errno));
  }

  zeroconf_svc_list = make_array(zeroconf_pool, 0, sizeof(DNSServiceRef));
  zeroconf_txt_list = make_array(zeroconf_pool, 0, sizeof(TXTRecordRef));

  zeroconf_register();
}

static void zeroconf_restart_ev(const void *event_data, void *user_data) {
  zeroconf_unregister();

  close(zeroconf_logfd);

  if (zeroconf_pool) {
    destroy_pool(zeroconf_pool);

    zeroconf_pool = make_sub_pool(permanent_pool);
    pr_pool_tag(zeroconf_pool, MOD_ZEROCONF_VERSION);

    zeroconf_svc_list = make_array(zeroconf_pool, 0,
      sizeof(DNSServiceRef));
    zeroconf_txt_list = make_array(zeroconf_pool, 0,
      sizeof(TXTRecordRef));
  }
}

/* Initialization functions
 */

static int zeroconf_init(void) {
  zeroconf_pool = make_sub_pool(permanent_pool);
  pr_pool_tag(zeroconf_pool, MOD_ZEROCONF_VERSION);

  pr_event_register(&zeroconf_module, "core.exit", zeroconf_exit_ev, NULL);
  pr_event_register(&zeroconf_module, "core.postparse", zeroconf_postparse_ev,
    NULL);
  pr_event_register(&zeroconf_module, "core.restart", zeroconf_restart_ev,
    NULL);

  return 0;
}

static int zeroconf_sess_init(void) {
  if (!zeroconf_engine)
    return 0;

  pr_event_unregister(&zeroconf_module, "core.exit", zeroconf_exit_ev);
  pr_event_unregister(&zeroconf_module, "core.postparse",
    zeroconf_postparse_ev);
  pr_event_unregister(&zeroconf_module, "core.restart", zeroconf_restart_ev);

  return 0;
}

/* Module API tables
 */

static conftable zeroconf_conftab[] = {
  { "ZeroconfEngine",	set_zeroconfengine,	NULL },
  { "ZeroconfInfo",	set_zeroconfinfo,	NULL },
  { "ZeroconfLog",	set_zeroconflog,	NULL },
  { NULL }
};

module zeroconf_module = {
  NULL, NULL,

  /* Module API version 2.0 */
  0x20,

  /* Module name */
  "zeroconf",

  /* Module configuration handler table */
  zeroconf_conftab,

  /* Module command handler table */
  NULL,

  /* Module authentication handler table */
  NULL,

  /* Module initialization function */
  zeroconf_init,

  /* Session initialization function */
  zeroconf_sess_init,

  /* Module version */
  MOD_ZEROCONF_VERSION
};
